__author__ = "Tomasz Rybotycki"

import unittest

from numpy import conjugate, identity, ndarray, transpose, allclose

from ..theboss.quantum_computations_utilities import generate_haar_random_unitary_matrix


class TestQuantumComputationsUtilities(unittest.TestCase):

    def setUp(self) -> None:
        self.matrix_size = 5
        self.number_of_matrices_for_distinct_elements_check = 10  # Should be >= 2.
        pass

    def test_unitarity_of_matrices_generated_by_haar_random_unitary_method(self) -> None:
        random_matrix = generate_haar_random_unitary_matrix(self.matrix_size)
        random_matrix_hermitian_adjoint = transpose(conjugate(random_matrix))
        product_of_matrix_and_hermitian_adjoint = random_matrix_hermitian_adjoint.dot(random_matrix)

        identity_matrix = identity(self.matrix_size, dtype=complex)

        self.assertTrue(self.__are_matrices_elementwise_close(identity_matrix, product_of_matrix_and_hermitian_adjoint))

    @staticmethod
    def __are_matrices_elementwise_close(matrix1: ndarray, matrix2: ndarray) -> bool:
        #  I assume that there are only rectangular matrices
        if len(matrix2) != len(matrix1):
            return False
        if len(matrix2[0]) != len(matrix1[0]):
            return False

        return allclose(matrix1, matrix2)

    def test_haar_random_unitary_matrices_generation_differences(self) -> None:
        generated_unitaries_matrices = \
            [generate_haar_random_unitary_matrix(self.matrix_size) for _ in
             range(self.number_of_matrices_for_distinct_elements_check)]

        are_all_matrices_different = []

        for i in range(self.number_of_matrices_for_distinct_elements_check):
            for j in range(i + 1, self.number_of_matrices_for_distinct_elements_check):
                are_all_matrices_different.append(self.__are_matrices_elementwise_close(generated_unitaries_matrices[i],
                                                                                        generated_unitaries_matrices[j])
                                                  )
        self.assertTrue(not any(are_all_matrices_different))



__author__ = "Tomasz Rybotycki"

import unittest
from copy import deepcopy
from random import uniform
from typing import List, Union

from numpy import asarray, block, eye, ndarray, zeros_like

from ..theboss.boson_sampling_simulator import BosonSamplingSimulator
from ..theboss.boson_sampling_utilities.permanent_calculators.bs_permanent_calculator_factory import \
    BSPermanentCalculatorFactory
from ..theboss.distribution_calculators.bs_exact_distribution_with_uniform_losses import \
    BSDistributionCalculatorWithUniformLosses, BosonSamplingExperimentConfiguration
from ..theboss.distribution_calculators.bs_sample_based_distribution_calculator import \
    BSSampleBasedDistributionCalculator
from ..theboss.network_simulation_strategy.lossy_network_simulation_strategy import \
    LossyNetworkSimulationStrategy
from ..theboss.quantum_computations_utilities import generate_haar_random_unitary_matrix, \
    count_total_variation_distance, count_tv_distance_error_bound_of_experiment_results
from ..theboss.simulation_strategies.simulation_strategy_factory import StrategyType, \
    SimulationStrategyFactory


class TestBosonSamplingClassicalSimulationStrategies(unittest.TestCase):

    def setUp(self) -> None:
        self._probability_of_uniform_loss = 0.8
        self._initial_state = [1, 1, 1, 1, 0, 0, 0, 0]
        self._number_of_samples_for_experiments = 1000
        self._probability_of_error_in_distribution_calculation = 0.001
        self._lossy_interferometer_matrix = self._probability_of_uniform_loss * generate_haar_random_unitary_matrix(
            len(self._initial_state))
        self._experiment_configuration = BosonSamplingExperimentConfiguration(
            interferometer_matrix=self._lossy_interferometer_matrix,
            initial_state=asarray(self._initial_state, dtype=int),
            initial_number_of_particles=sum(self._initial_state),
            number_of_modes=len(self._initial_state),
            number_of_particles_lost=0,  # Losses should only come from network.
            number_of_particles_left=sum(self._initial_state),
            uniform_transmissivity=self._probability_of_uniform_loss,
            network_simulation_strategy=LossyNetworkSimulationStrategy(
                self._lossy_interferometer_matrix)
        )

        permanent_calculator_factory = BSPermanentCalculatorFactory(
            self._lossy_interferometer_matrix, None, None)
        self._permanent_calculator = permanent_calculator_factory.generate_calculator()
        self._strategy_factory = SimulationStrategyFactory(
            self._experiment_configuration, self._permanent_calculator)

        calculator = BSDistributionCalculatorWithUniformLosses(
            self._experiment_configuration,
            self._permanent_calculator)
        self._possible_outcomes = calculator.get_outcomes_in_proper_order()
        self._possible_outcomes_number = len(self._possible_outcomes)

    def test_lossy_network_simulation_number_of_particles(self) -> None:
        """
        This test checks if number of particles for lossy network simulator is lower than number of
        particles without the losses. Note that I can be pretty sure it will be lower in large losses
        regime, but in case of lower losses this test may not hold.
        """

        self._strategy_factory.strategy_type = StrategyType.FIXED_LOSS

        strategy = self._strategy_factory.generate_strategy()
        simulator = BosonSamplingSimulator(strategy)
        lossy_average_number_of_particles = 0
        samples = simulator.get_classical_simulation_results(
            asarray(self._initial_state, dtype=int),
            self._number_of_samples_for_experiments)
        for sample in samples:
            lossy_average_number_of_particles += sum(sample)

        lossy_average_number_of_particles /= self._number_of_samples_for_experiments
        lossless_number_of_particles = sum(self._initial_state)
        self.assertLess(lossy_average_number_of_particles, lossless_number_of_particles)

    def test_lossy_network_with_uniform_losses_distribution_accuracy(self) -> None:
        """
        This test is for uniform losses. It checks if uniform losses in network simulations are close to
        simulation with uniform losses at the input. Given that both approximate the same quantity only
        sampling complexity error is assumed.
        """
        self._strategy_factory.strategy_type = StrategyType.UNIFORM_LOSS

        estimated_distribution_calculator = BSSampleBasedDistributionCalculator(
            experiment_configuration=self._experiment_configuration,
            strategy=self._strategy_factory.generate_strategy(),
            outcomes=self._possible_outcomes
        )

        uniform_losses_distribution = estimated_distribution_calculator.calculate_approximate_distribution()

        self.__check_if_given_distribution_is_close_to_lossy_network_distribution(
            uniform_losses_distribution)

    def __check_if_given_distribution_is_close_to_lossy_network_distribution(
            self, distribution: Union[ndarray, List[float]]) -> None:
        """
        Common part of distribution comparison tests.
        :param distribution: Given distribution to compare with lossy distribution.
        """

        self._strategy_factory.strategy_type = StrategyType.FIXED_LOSS

        distance_bound_between_estimated_distributions = \
            self.__calculate_statistical_distance_bound_between_two_approximate_distributions(
                outcomes_number=self._possible_outcomes_number)

        estimated_distribution_calculator = BSSampleBasedDistributionCalculator(
            experiment_configuration=self._experiment_configuration,
            strategy=self._strategy_factory.generate_strategy(),
            outcomes=self._possible_outcomes
        )

        lossy_network_distribution = estimated_distribution_calculator.calculate_approximate_distribution()

        distance_between_distributions = count_total_variation_distance(
            lossy_network_distribution, distribution
        )

        self.assertLessEqual(distance_between_distributions,
                             distance_bound_between_estimated_distributions)

    def __calculate_statistical_distance_bound_between_two_approximate_distributions(
            self,
            outcomes_number: int) -> float:
        return 2 * count_tv_distance_error_bound_of_experiment_results(
            outcomes_number=outcomes_number,
            samples_number=self._number_of_samples_for_experiments,
            error_probability=self._probability_of_error_in_distribution_calculation
        )

    def test_lossy_network_with_uniform_losses_distribution_accuracy_against_generalized_cliffords(
            self) -> None:
        """
        This test checks if lossy networks distribution provides same approximate (up to statistical bounds)
        distribution as generalized cliffords distribution with lossy inputs.
        """
        generalized_cliffords_distribution = self.__calculate_generalized_cliffords_distribution_with_lossy_inputs()

        self.__check_if_given_distribution_is_close_to_lossy_network_distribution(
            generalized_cliffords_distribution)

    def __calculate_generalized_cliffords_distribution_with_lossy_inputs(self) -> List[
        float]:
        """
        This method calculates approximate distribution for lossy states using generalized cliffords
        method.
        :return: Approximate distribution.
        """
        self._strategy_factory.strategy_type = StrategyType.GENERALIZED_CLIFFORD
        self._permanent_calculator.matrix = self._lossy_interferometer_matrix
        strategy = self._strategy_factory.generate_strategy()
        simulator = BosonSamplingSimulator(strategy)

        samples = simulator.get_classical_simulation_results(
            self.__get_uniformly_lossy_input_state(),
            self._number_of_samples_for_experiments)

        return self.__calculate_distribution(samples, self._possible_outcomes)

    @staticmethod
    def __calculate_distribution(samples: List[ndarray],
                                 possible_outcomes: List[ndarray]) -> List[float]:
        probabilities = [0] * len(possible_outcomes)

        for sample in samples:
            for i in range(len(possible_outcomes)):
                # Check if obtained result is one of possible outcomes.
                if all(sample == possible_outcomes[
                    i]):  # Expect all elements of resultant list to be True.
                    probabilities[i] += 1
                    break

        for i in range(len(probabilities)):
            probabilities[i] /= len(samples)

        return probabilities

    def __get_uniformly_lossy_input_state(self) -> ndarray:
        """
        This method assumes that losses are uniform and specified in the configuration of the
        experiment (in test case setup).
        :return: Input state after losses.
        """
        lossy_input = asarray(self._initial_state, dtype=int)
        for i in range(len(self._initial_state)):
            for _ in range(self._initial_state[i]):
                if uniform(0,
                           1) < self._experiment_configuration.uniform_transmissivity:
                    lossy_input[i] -= 1
        return lossy_input

    def test_lossy_network_with_uniform_losses_on_bosonful_modes_and_higher_losses_on_bosonless(
            self) -> None:
        """
        This test checks the situation, for which losses are uniform only for these modes that have any bosons in it.
        It assumes that initial state is [1, 1, 1, 1, 0, 0, 0, 0] as in initial version of the tests case. It goes to
        extreme making bosonless modes completely lossy. The results should be identical (up to statistical error).
        """
        experiment_configuration = deepcopy(self._experiment_configuration)
        updated_interferometer_matrix = experiment_configuration.interferometer_matrix
        identity_on_bosonful_modes = eye(sum(self._initial_state), dtype=complex)
        zeros_on_bosonless_modes = zeros_like(identity_on_bosonful_modes, dtype=complex)
        update_matrix = block([
            [identity_on_bosonful_modes, zeros_on_bosonless_modes],
            [zeros_on_bosonless_modes, zeros_on_bosonless_modes]
        ])
        updated_interferometer_matrix = update_matrix @ updated_interferometer_matrix

        experiment_configuration.interferometer_matrix = updated_interferometer_matrix

        self._strategy_factory.strategy_type = StrategyType.FIXED_LOSS
        self._strategy_factory.experiment_configuration = experiment_configuration

        estimated_distribution_calculator = BSSampleBasedDistributionCalculator(
            experiment_configuration=self._experiment_configuration,
            strategy=self._strategy_factory.generate_strategy(),
            outcomes=self._possible_outcomes
        )

        distribution_with_huge_losses_on_bosonless_modes = \
            estimated_distribution_calculator.calculate_approximate_distribution()

        self._strategy_factory.experiment_configuration = self._experiment_configuration
        self.__check_if_given_distribution_is_close_to_lossy_network_distribution(
            distribution_with_huge_losses_on_bosonless_modes)

    def test_distance_of_generalized_cliffords_with_lossy_network_and_lossy_input(
            self) -> None:
        distribution_with_lossy_net = self.__calculate_generalized_cliffords_distribution_with_lossy_network()
        distribution_with_lossy_input = self.__calculate_generalized_cliffords_distribution_with_lossy_inputs()

        distance_bound_between_estimated_distributions = \
            self.__calculate_statistical_distance_bound_between_two_approximate_distributions(
                outcomes_number=self._possible_outcomes_number)

        distance_between_distributions = count_total_variation_distance(
            distribution_with_lossy_input, distribution_with_lossy_net
        )

        self.assertLessEqual(distance_between_distributions,
                             distance_bound_between_estimated_distributions)

    def __calculate_generalized_cliffords_distribution_with_lossy_network(self) -> List[
        float]:
        """
        This method calculates approximate distribution for lossy states using generalized cliffords
        method.
        :return: Approximate distribution.
        """
        self._strategy_factory.strategy_type = StrategyType.LOSSY_NET_GENERALIZED_CLIFFORD
        strategy = self._strategy_factory.generate_strategy()
        simulator = BosonSamplingSimulator(strategy)
        samples = simulator.get_classical_simulation_results(
            asarray(self._initial_state, dtype=int),
            self._number_of_samples_for_experiments)

        return self.__calculate_distribution(samples, self._possible_outcomes)



__author__ = "Tomasz Rybotycki"

"""
    The intention behind this code is to have tests for different simulation strategies in one place. The reason for
    that is that their code stays mostly the same and with every different strategy I end up writing more and more of
    the same stuff with only minor differences (mostly in experiments setup). 
"""

import unittest
from copy import deepcopy
from dataclasses import dataclass
from math import factorial
from typing import List

from numpy import array, average, ndarray
from numpy.random import randint
from scipy.special import binom

from ..theboss.simulation_strategies.simulation_strategy_factory import StrategyType, SimulationStrategyFactory, SimulationStrategyInterface

from ..theboss.boson_sampling_utilities.boson_sampling_utilities import calculate_number_of_possible_lossy_n_particle_m_mode_output_states, calculate_number_of_possible_n_particle_m_mode_output_states
from ..theboss.quantum_computations_utilities import generate_haar_random_unitary_matrix, count_total_variation_distance, count_tv_distance_error_bound_of_experiment_results
from ..theboss.network_simulation_strategy.lossless_network_simulation_strategy import LosslessNetworkSimulationStrategy

from ..theboss.boson_sampling_utilities.permanent_calculators.bs_permanent_calculator_factory import BSPermanentCalculatorFactory
from ..theboss.distribution_calculators.bs_exact_distribution_with_uniform_losses import BSDistributionCalculatorWithUniformLosses, BSDistributionCalculatorWithFixedLosses, BosonSamplingExperimentConfiguration
from ..theboss.distribution_calculators.bs_sample_based_distribution_calculator import BSSampleBasedDistributionCalculator, BSDistributionCalculatorInterface


# TODO TR: These tests has to be though through and refactored.
# TODO TR: Add tests for lossy state approximation strategy

@dataclass
class DistributionAccuracyExperimentConfiguration:
    exact_calculator: BSDistributionCalculatorInterface
    estimation_calculator: BSSampleBasedDistributionCalculator
    approximation_distance_bound: float
    outcomes_number: int


class TestBosonSamplingClassicalSimulationStrategies(unittest.TestCase):

    def setUp(self) -> None:
        print(f"\nIn method {self._testMethodName}. Test start!\n")
        self._permutation_matrix = array([
            [0, 0, 1, 0, 0],
            [1, 0, 0, 0, 0],
            [0, 0, 0, 1, 0],
            [0, 0, 0, 0, 1],
            [0, 1, 0, 0, 0],
        ], dtype=complex)

        self._number_of_samples_for_estimated_distribution_calculation = int(1e3)
        self._probability_of_error_in_distribution_calculation = 1e-3

        self._distance_calculation_initial_state = [1, 1, 1, 1, 0]
        self._distance_calculation_binned_initial_state = [2, 1, 1, 0, 0]
        self._distance_calculation_number_of_particles_lost = 2
        self._uniform_transmissivity = 0.8

        self._nonuniform_strategy_initial_state = [0, 2, 1, 1, 0]
        self._approximated_modes_number = 2

        distance_calculation_initial_number_of_particles = sum(self._distance_calculation_initial_state)

        self._distributions_distance_experiment_configuration = BosonSamplingExperimentConfiguration(
            interferometer_matrix=self._permutation_matrix,
            initial_state=array(self._distance_calculation_initial_state, dtype=int),
            initial_number_of_particles=distance_calculation_initial_number_of_particles,
            number_of_modes=len(self._distance_calculation_initial_state),
            number_of_particles_lost=self._distance_calculation_number_of_particles_lost,
            number_of_particles_left=distance_calculation_initial_number_of_particles -
                                     self._distance_calculation_number_of_particles_lost,
            uniform_transmissivity=0.8
        )

        permanent_calculator_factory = BSPermanentCalculatorFactory(self._permutation_matrix, None, None)
        self._bs_permanent_calculator = permanent_calculator_factory.generate_calculator()

        self._strategies_factory = SimulationStrategyFactory(self._distributions_distance_experiment_configuration,
                                                             self._bs_permanent_calculator)

        self._haar_random_matrices_number = 10
        self._haar_random_experiment_input_state = [1, 1, 1, 1, 0]
        haar_random_initial_number_of_particles = sum(self._haar_random_experiment_input_state)
        haar_random_number_of_particles_lost = 2
        self._haar_random_experiment_configuration = BosonSamplingExperimentConfiguration(
            interferometer_matrix=array([], dtype=complex),
            initial_state=array(self._haar_random_experiment_input_state, dtype=int),
            initial_number_of_particles=haar_random_number_of_particles_lost,
            number_of_modes=len(self._haar_random_experiment_input_state),
            number_of_particles_lost=haar_random_number_of_particles_lost,
            number_of_particles_left=haar_random_initial_number_of_particles - haar_random_number_of_particles_lost
        )
        self._haar_random_binned_experiment_input_state = [2, 1, 1, 1, 0]

    def tearDown(self) -> None:
        print("\nTest finished!\n")

    def test_distribution_accuracy_for_fixed_losses_strategy(self) -> None:
        self.__prepare_lossy_distance_experiment_settings()
        self._distributions_distance_experiment_configuration.network_simulation_strategy = \
            LosslessNetworkSimulationStrategy(self._permutation_matrix)

        self._strategies_factory.experiment_configuration = self._distributions_distance_experiment_configuration
        self._strategies_factory.strategy_type = StrategyType.FIXED_LOSS

        distance_experiment_configuration = DistributionAccuracyExperimentConfiguration(
            exact_calculator=BSDistributionCalculatorWithFixedLosses(
                self._distributions_distance_experiment_configuration,
                self._bs_permanent_calculator),
            estimation_calculator=self.__generate_estimated_distribution_calculator(
                self._strategies_factory.generate_strategy()
            ),
            outcomes_number=calculate_number_of_possible_n_particle_m_mode_output_states(
                n=self._distributions_distance_experiment_configuration.number_of_particles_left,
                m=self._distributions_distance_experiment_configuration.number_of_modes
            ),
            approximation_distance_bound=self.__calculate_fixed_losses_distance_bound_from_exact_to_experimental(
                self._distributions_distance_experiment_configuration.initial_number_of_particles,
                self._distributions_distance_experiment_configuration.number_of_particles_left
            )
        )
        self.__perform_distance_of_approximate_distribution_from_ideal(distance_experiment_configuration)

    def __prepare_lossy_distance_experiment_settings(self) -> None:
        # At least for now lossy experiments are only for classic [1, 1, ..., 1, 0, 0, ..., 0] states.
        self._distributions_distance_experiment_configuration.initial_state = self._distance_calculation_initial_state
        initial_number_of_particles = sum(self._distance_calculation_initial_state)
        self._distributions_distance_experiment_configuration.initial_number_of_particles = initial_number_of_particles
        self._distributions_distance_experiment_configuration.number_of_particles_lost = \
            self._distance_calculation_number_of_particles_lost
        self._distributions_distance_experiment_configuration.number_of_particles_left = \
            initial_number_of_particles - self._distance_calculation_number_of_particles_lost
        self._distributions_distance_experiment_configuration.uniform_transmissivity = self._uniform_transmissivity

    def __perform_distance_of_approximate_distribution_from_ideal(
            self, distance_accuracy_experiment_configuration: DistributionAccuracyExperimentConfiguration) -> None:
        distance_from_exact_to_estimated = self.__calculate_distance_from_exact_distribution_to_estimated_distribution(
            exact_distribution_calculator=distance_accuracy_experiment_configuration.exact_calculator,
            estimated_distribution_calculator=distance_accuracy_experiment_configuration.estimation_calculator
        )

        distance_from_approximation_to_estimated = count_tv_distance_error_bound_of_experiment_results(
            outcomes_number=distance_accuracy_experiment_configuration.outcomes_number,
            samples_number=self._number_of_samples_for_estimated_distribution_calculation,
            error_probability=self._probability_of_error_in_distribution_calculation
        )

        # Using triangle inequality of (TV) distance.
        max_allowed_distance = \
            distance_accuracy_experiment_configuration.approximation_distance_bound \
            + distance_from_approximation_to_estimated

        self.assertLessEqual(distance_from_exact_to_estimated, max_allowed_distance,
                             f"Distance from exact distribution ({distance_from_exact_to_estimated}) is "
                             f"greater than maximum distance allowed ({max_allowed_distance}).")

    def __generate_estimated_distribution_calculator(self, strategy: SimulationStrategyInterface,
                                                     outcomes: List[ndarray] = None) \
            -> BSSampleBasedDistributionCalculator:
        estimated_distribution_calculator = BSSampleBasedDistributionCalculator(
            experiment_configuration=self._distributions_distance_experiment_configuration,
            strategy=strategy,
            outcomes=outcomes
        )
        return estimated_distribution_calculator

    def __calculate_distance_from_exact_distribution_to_estimated_distribution(
            self, exact_distribution_calculator: BSDistributionCalculatorInterface,
            estimated_distribution_calculator: BSSampleBasedDistributionCalculator) -> float:
        """
            Using specified calculators, it counts exact and estimated probability distributions and calculates
            the tv distance between them.
            :param exact_distribution_calculator: Calculator of the exact (ideal, bruteforce permanents) distribution.
            :param estimated_distribution_calculator: Calculator of the approximate (sampling according to strategy)
            distribution.
            :return: TV distance between exact and approximated distribution.
        """
        exact_distribution = exact_distribution_calculator.calculate_distribution()
        approximated_distribution = estimated_distribution_calculator.calculate_approximate_distribution(
            samples_number=self._number_of_samples_for_estimated_distribution_calculation
        )
        return count_total_variation_distance(exact_distribution, approximated_distribution)

    @staticmethod
    def __calculate_fixed_losses_distance_bound_from_exact_to_experimental(initial_number_of_particles: int,
                                                                           number_of_particles_left: int) -> float:
        """
            This is the distance bound between experimental and ideal results for fixed losses boson sampling.
            Basically this is capital Delta from [1]. Using n and l notation from the paper [1] for readability
            purposes.
            This will be used for uniform losses and fixed losses bound calculation.
            :return: Distance bound between experimental and ideal results for uniform losses.
        """
        n = initial_number_of_particles
        l = number_of_particles_left
        error_bound = 1.0 - (factorial(n) / (pow(n, l) * factorial(n - l)))
        return error_bound

    def test_distribution_accuracy_for_generalized_cliffords_strategy(self) -> None:
        self.__prepare_lossless_distance_experiment_settings()
        self.__continue_with_common_part_of_lossless_generalized_cliffords_based_strategy_tests()

    def __prepare_lossless_distance_experiment_settings(self) -> None:
        self._distributions_distance_experiment_configuration.initial_state = self._distance_calculation_initial_state
        initial_number_of_particles = sum(self._distance_calculation_initial_state)
        self._distributions_distance_experiment_configuration.initial_number_of_particles = initial_number_of_particles
        self._distributions_distance_experiment_configuration.number_of_particles_lost = 0
        self._distributions_distance_experiment_configuration.number_of_particles_left = initial_number_of_particles
        self._distributions_distance_experiment_configuration.uniform_transmissivity = 1

    def __continue_with_common_part_of_lossless_generalized_cliffords_based_strategy_tests(self,
                                                                                           strategy: StrategyType = StrategyType.GENERALIZED_CLIFFORD) -> None:
        self._strategies_factory.experiment_configuration = self._distributions_distance_experiment_configuration
        self._strategies_factory.strategy_type = strategy
        distance_experiment_configuration = DistributionAccuracyExperimentConfiguration(
            # This exact calculator, when there are no losses, will do the work just fine.
            exact_calculator=BSDistributionCalculatorWithFixedLosses(
                self._distributions_distance_experiment_configuration,
                self._bs_permanent_calculator),
            estimation_calculator=self.__generate_estimated_distribution_calculator(
                self._strategies_factory.generate_strategy()
            ),
            outcomes_number=calculate_number_of_possible_n_particle_m_mode_output_states(
                n=self._distributions_distance_experiment_configuration.number_of_particles_left,
                m=self._distributions_distance_experiment_configuration.number_of_modes
            ),
            approximation_distance_bound=0  # This strategy returns exact solution.
        )
        self.__perform_distance_of_approximate_distribution_from_ideal(distance_experiment_configuration)

    def test_distribution_accuracy_for_generalized_cliffords_strategy_with_binned_input(self) -> None:
        self.__prepare_lossless_distance_experiments_settings_with_binned_inputs()
        self.__continue_with_common_part_of_lossless_generalized_cliffords_based_strategy_tests()

    def __prepare_lossless_distance_experiments_settings_with_binned_inputs(self) -> None:
        self._distributions_distance_experiment_configuration.initial_state = \
            self._distance_calculation_binned_initial_state
        initial_number_of_particles = sum(self._distance_calculation_binned_initial_state)
        self._distributions_distance_experiment_configuration.initial_number_of_particles = initial_number_of_particles
        self._distributions_distance_experiment_configuration.number_of_particles_lost = 0
        self._distributions_distance_experiment_configuration.number_of_particles_left = initial_number_of_particles
        self._distributions_distance_experiment_configuration.uniform_transmissivity = 1

    def test_distribution_accuracy_for_lossy_network_generalized_cliffords_strategy_without_losses(self):
        self.__prepare_lossless_distance_experiment_settings()
        self.__continue_with_common_part_of_lossless_generalized_cliffords_based_strategy_tests(
            StrategyType.LOSSY_NET_GENERALIZED_CLIFFORD)

    def test_distribution_accuracy_for_lossy_network_generalized_cliffords_strategy_without_losses_with_binned_input(
            self):
        self.__prepare_lossless_distance_experiments_settings_with_binned_inputs()
        self.__continue_with_common_part_of_lossless_generalized_cliffords_based_strategy_tests(
            StrategyType.LOSSY_NET_GENERALIZED_CLIFFORD)

    def test_distribution_accuracy_for_uniform_losses_generalized_cliffords_without_losses(self):
        self.__prepare_lossless_distance_experiment_settings()
        self.__continue_with_common_part_of_lossless_generalized_cliffords_based_strategy_tests(
            StrategyType.GENERALIZED_U_LOSSY_CLIFFORD)

    def test_distribution_accuracy_for_uniform_losses_generalized_cliffords_without_losses_with_binned_input(self):
        self.__prepare_lossless_distance_experiments_settings_with_binned_inputs()
        self.__continue_with_common_part_of_lossless_generalized_cliffords_based_strategy_tests(
            StrategyType.GENERALIZED_U_LOSSY_CLIFFORD)

    def test_distribution_accuracy_for_nonuniform_losses_approximated_strategy_without_approximated_modes_or_loses(self):
        self.__prepare_lossless_distance_experiment_settings()
        self._distributions_distance_experiment_configuration.hierarchy_level = self._distributions_distance_experiment_configuration.number_of_modes
        self.__continue_with_common_part_of_lossless_generalized_cliffords_based_strategy_tests(
            StrategyType.NONUNIFORM_APPROXIMATED)

    def test_distribution_accuracy_for_nonuniform_losses_approximated_strategy_without_approximated_modes_or_loses_with_binned_input(self):
        self.__prepare_lossless_distance_experiments_settings_with_binned_inputs()
        self._distributions_distance_experiment_configuration.hierarchy_level = self._distributions_distance_experiment_configuration.number_of_modes
        self.__continue_with_common_part_of_lossless_generalized_cliffords_based_strategy_tests(
            StrategyType.NONUNIFORM_APPROXIMATED)

    def test_distribution_accuracy_for_uniform_losses_strategy(self) -> None:
        self.__prepare_lossy_distance_experiment_settings()
        self._strategies_factory.experiment_configuration = self._distributions_distance_experiment_configuration
        self._strategies_factory.strategy_type = StrategyType.UNIFORM_LOSS
        exact_calculator = BSDistributionCalculatorWithUniformLosses(
            self._distributions_distance_experiment_configuration, self._bs_permanent_calculator)
        distance_experiment_configuration = DistributionAccuracyExperimentConfiguration(
            exact_calculator=exact_calculator,
            estimation_calculator=self.__generate_estimated_distribution_calculator(
                self._strategies_factory.generate_strategy(),
                outcomes=exact_calculator.get_outcomes_in_proper_order()
            ),
            outcomes_number=calculate_number_of_possible_lossy_n_particle_m_mode_output_states(
                n=self._distributions_distance_experiment_configuration.number_of_particles_left,
                m=self._distributions_distance_experiment_configuration.number_of_modes
            ),
            approximation_distance_bound=self.__calculate_uniform_loss_distribution_error_bound(),
        )
        self.__perform_distance_of_approximate_distribution_from_ideal(distance_experiment_configuration)

    def __calculate_uniform_loss_distribution_error_bound(self) -> float:
        """
            This is the distance bound between approximated and ideal results for uniform losses boson sampling.
            Basically this is capital Delta from [1]. Using eta, n and l notation from the paper [1] for readability
            purposes.
        :return: Distance bound between experimental and ideal results for uniform losses.
        """
        #
        error_bound = 0
        n = self._distributions_distance_experiment_configuration.initial_number_of_particles
        eta = self._distributions_distance_experiment_configuration.uniform_transmissivity
        for number_of_particles_left in range(n + 1):
            l = number_of_particles_left
            subdistribution_weight = pow(eta, l) * binom(n, l) * pow(1.0 - eta, n - l)
            error_bound += \
                subdistribution_weight * self.__calculate_fixed_losses_distance_bound_from_exact_to_experimental(n, l)
        return error_bound

    def test_distribution_accuracy_for_u_lossy_generalized_cliffords_strategy_with_losses(self):
        self.__prepare_lossy_distance_experiment_settings()
        self.__continue_with_common_part_of_uniformly_lossy_generalized_cliffords_based_strategy_tests(StrategyType.GENERALIZED_U_LOSSY_CLIFFORD)

    def __continue_with_common_part_of_uniformly_lossy_generalized_cliffords_based_strategy_tests(self,
                                                                                                  strategy: StrategyType = StrategyType.GENERALIZED_U_LOSSY_CLIFFORD) -> None:
        self._strategies_factory.experiment_configuration = self._distributions_distance_experiment_configuration
        self._strategies_factory.strategy_type = strategy
        exact_calculator = BSDistributionCalculatorWithUniformLosses(
            self._distributions_distance_experiment_configuration, self._bs_permanent_calculator)

        if strategy == StrategyType.LOSSY_NET_GENERALIZED_CLIFFORD or strategy == StrategyType.NONUNIFORM_APPROXIMATED:
            self._strategies_factory.bs_permanent_calculator.matrix *= pow(self._uniform_transmissivity, 0.5)

        distance_experiment_configuration = DistributionAccuracyExperimentConfiguration(
            # This exact calculator, when there are no losses, will do the work just fine.
            exact_calculator=exact_calculator,
            estimation_calculator=self.__generate_estimated_distribution_calculator(
                self._strategies_factory.generate_strategy(),
                outcomes=exact_calculator.get_outcomes_in_proper_order()
            ),
            outcomes_number=calculate_number_of_possible_lossy_n_particle_m_mode_output_states(
                n=self._distributions_distance_experiment_configuration.number_of_particles_left,
                m=self._distributions_distance_experiment_configuration.number_of_modes
            ),
            approximation_distance_bound=0  # This strategy returns exact solution.
        )
        self.__perform_distance_of_approximate_distribution_from_ideal(distance_experiment_configuration)

    def test_distribution_accuracy_for_net_lossy_generalized_cliffords_strategy_with_losses(self):
        self.__prepare_lossy_distance_experiment_settings()
        self.__continue_with_common_part_of_uniformly_lossy_generalized_cliffords_based_strategy_tests(
            StrategyType.LOSSY_NET_GENERALIZED_CLIFFORD)

    def test_distribution_accuracy_for_u_lossy_generalized_cliffords_strategy_with_losses_and_binned_input(self):
        self.__prepare_lossy_distance_experiment_settings_with_binned_input()
        self.__continue_with_common_part_of_uniformly_lossy_generalized_cliffords_based_strategy_tests(
            StrategyType.GENERALIZED_U_LOSSY_CLIFFORD)

    def __prepare_lossy_distance_experiment_settings_with_binned_input(self):
        # At least for now lossy experiments are only for classic [1, 1, ..., 1, 0, 0, ..., 0] states.
        self._distributions_distance_experiment_configuration.initial_state = self._distance_calculation_binned_initial_state
        initial_number_of_particles = sum(self._distance_calculation_initial_state)
        self._distributions_distance_experiment_configuration.initial_number_of_particles = initial_number_of_particles
        self._distributions_distance_experiment_configuration.number_of_particles_lost = \
            self._distance_calculation_number_of_particles_lost
        self._distributions_distance_experiment_configuration.number_of_particles_left = \
            initial_number_of_particles - self._distance_calculation_number_of_particles_lost
        self._distributions_distance_experiment_configuration.uniform_transmissivity = self._uniform_transmissivity

    def test_distribution_accuracy_for_net_lossy_generalized_cliffords_strategy_with_losses_and_binned_input(self):
        self.__prepare_lossy_distance_experiment_settings_with_binned_input()
        self.__continue_with_common_part_of_uniformly_lossy_generalized_cliffords_based_strategy_tests(
            StrategyType.LOSSY_NET_GENERALIZED_CLIFFORD)

    def test_distribution_accuracy_for_nonuniform_losses_approximated_strategy_without_approximated_modes(self):
        self.__prepare_lossy_distance_experiment_settings()
        self._distributions_distance_experiment_configuration.hierarchy_level = self._distributions_distance_experiment_configuration.number_of_modes
        self.__continue_with_common_part_of_uniformly_lossy_generalized_cliffords_based_strategy_tests(StrategyType.NONUNIFORM_APPROXIMATED)

    def test_distribution_accuracy_for_nonuniform_losses_approximated_strategy_without_approximated_modes_with_binned_input(self):
        self.__prepare_lossy_distance_experiment_settings_with_binned_input()
        self._distributions_distance_experiment_configuration.hierarchy_level = self._distributions_distance_experiment_configuration.number_of_modes
        self.__continue_with_common_part_of_uniformly_lossy_generalized_cliffords_based_strategy_tests(StrategyType.NONUNIFORM_APPROXIMATED)

    def test_distribution_accuracy_for_nonuniform_losses_approximated_strategy_with_approximated_modes(self):
        self.__prepare_lossy_distance_experiment_settings()
        self._distributions_distance_experiment_configuration.approximated_modes_number = \
            self._approximated_modes_number
        self.__continue_with_common_part_of_uniformly_lossy_nonuniform_losses_approximated_strategy_tests()

    def __continue_with_common_part_of_uniformly_lossy_nonuniform_losses_approximated_strategy_tests(self) -> None:

        self._strategies_factory.experiment_configuration = self._distributions_distance_experiment_configuration
        self._strategies_factory.strategy_type = StrategyType.NONUNIFORM_APPROXIMATED
        exact_calculator = BSDistributionCalculatorWithUniformLosses(
            self._distributions_distance_experiment_configuration, self._bs_permanent_calculator)

        self._distributions_distance_experiment_configuration.initial_state = \
            self._nonuniform_strategy_initial_state
        self._strategies_factory.bs_permanent_calculator.matrix *= pow(self._uniform_transmissivity, 0.5)

        distance_experiment_configuration = DistributionAccuracyExperimentConfiguration(
            # This exact calculator, when there are no losses, will do the work just fine.
            exact_calculator=exact_calculator,
            estimation_calculator=self.__generate_estimated_distribution_calculator(
                self._strategies_factory.generate_strategy(),
                outcomes=exact_calculator.get_outcomes_in_proper_order()
            ),
            outcomes_number=calculate_number_of_possible_lossy_n_particle_m_mode_output_states(
                n=self._distributions_distance_experiment_configuration.number_of_particles_left,
                m=self._distributions_distance_experiment_configuration.number_of_modes
            ),
            approximation_distance_bound=self.__calculate_nonuniform_losses_distance_bound_from_exact_to_experimental()
        )
        self.__perform_distance_of_approximate_distribution_from_ideal(distance_experiment_configuration)

    def __calculate_nonuniform_losses_distance_bound_from_exact_to_experimental(self):

        eta_eff = self._uniform_transmissivity

        bound = pow(eta_eff, 2) / 2
        bound *= (self._distributions_distance_experiment_configuration.initial_number_of_particles -
                  self._distributions_distance_experiment_configuration.approximated_modes_number)
        bound += eta_eff * (1 - eta_eff) / 2
        return bound

    def test_haar_random_interferometers_distance_for_fixed_losses_strategy(self) -> None:
        self.__set_experiment_configuration_for_standard_haar_random()
        strategy_factory = SimulationStrategyFactory(self._haar_random_experiment_configuration,
                                                     self._bs_permanent_calculator,
                                                     StrategyType.FIXED_LOSS)
        self.__test_haar_random_interferometers_approximation_distance_from_ideal(strategy_factory)

    def test_haar_random_interferometers_distance_for_cliffords_r_strategy(self) -> None:
        self.__set_experiment_configuration_for_lossless_haar_random()
        try:
            strategy_factory = SimulationStrategyFactory(self._haar_random_experiment_configuration,
                                                         self._bs_permanent_calculator,
                                                         StrategyType.CLIFFORD_R)
            self.__test_haar_random_interferometers_approximation_distance_from_ideal(
                strategy_factory)
        except Exception as e:
            print("You don't have packages required for this strategy.")
            print(e)
            self.assertTrue(True)  # Maybe assertion should be true?


    def test_haar_random_interferometers_distance_for_generalized_cliffords_strategy(self) -> None:
        self.__set_experiment_configuration_for_lossless_haar_random()
        strategy_factory = SimulationStrategyFactory(self._haar_random_experiment_configuration,
                                                     self._bs_permanent_calculator,
                                                     StrategyType.GENERALIZED_CLIFFORD)
        self.__test_haar_random_interferometers_approximation_distance_from_ideal(strategy_factory)

    def test_haar_random_interferometers_distance_for_generalized_cliffords_strategy_with_binned_input(self) -> None:
        self.__set_experiment_configuration_for_binned_haar_random()
        strategy_factory = SimulationStrategyFactory(self._haar_random_experiment_configuration,
                                                     self._bs_permanent_calculator,
                                                     StrategyType.GENERALIZED_CLIFFORD)
        self.__test_haar_random_interferometers_approximation_distance_from_ideal(strategy_factory)

    def __set_experiment_configuration_for_binned_haar_random(self) -> None:
        # Note that binned configuration is also for lossless case (for now)
        self._haar_random_experiment_configuration.initial_state = self._haar_random_binned_experiment_input_state
        number_of_particles_in_the_experiment = len(self._haar_random_binned_experiment_input_state)
        self._haar_random_experiment_configuration.initial_number_of_particles = number_of_particles_in_the_experiment
        self._haar_random_experiment_configuration.number_of_particles_left = number_of_particles_in_the_experiment

    def __set_experiment_configuration_for_standard_haar_random(self) -> None:
        self._haar_random_experiment_configuration.initial_state = self._haar_random_experiment_input_state
        number_of_particles_in_the_experiment = sum(self._haar_random_experiment_input_state)
        self._haar_random_experiment_configuration.initial_number_of_particles = number_of_particles_in_the_experiment
        self._haar_random_experiment_configuration.number_of_particles_left = \
            number_of_particles_in_the_experiment - self._haar_random_experiment_configuration.number_of_particles_lost

    def __set_experiment_configuration_for_lossless_haar_random(self) -> None:
        self._haar_random_experiment_configuration.initial_state = self._haar_random_experiment_input_state
        number_of_particles_in_the_experiment = sum(self._haar_random_experiment_input_state)
        self._haar_random_experiment_configuration.initial_number_of_particles = number_of_particles_in_the_experiment
        self._haar_random_experiment_configuration.number_of_particles_left = number_of_particles_in_the_experiment

    def __test_haar_random_interferometers_approximation_distance_from_ideal(
            self, strategy_factory: SimulationStrategyFactory) -> None:
        number_of_outcomes = calculate_number_of_possible_n_particle_m_mode_output_states(
            n=self._haar_random_experiment_configuration.number_of_particles_left,
            m=self._haar_random_experiment_configuration.number_of_modes
        )

        error_bound = count_tv_distance_error_bound_of_experiment_results(
            outcomes_number=number_of_outcomes, samples_number=self._haar_random_matrices_number,
            error_probability=self._probability_of_error_in_distribution_calculation
        )

        probabilities_list = []
        current_probabilities = []

        for i in range(self._haar_random_matrices_number):

            print(f"Current Haar random matrix index: {i} out of {self._haar_random_matrices_number}.")

            experiment_configuration = deepcopy(self._haar_random_experiment_configuration)
            experiment_configuration.interferometer_matrix = generate_haar_random_unitary_matrix(
                self._haar_random_experiment_configuration.number_of_modes)
            strategy_factory.experiment_configuration = experiment_configuration
            self._bs_permanent_calculator.matrix = experiment_configuration.interferometer_matrix
            strategy_factory.bs_permanent_calculator = self._bs_permanent_calculator
            distribution_calculator = BSSampleBasedDistributionCalculator(experiment_configuration,
                                                                          strategy_factory.generate_strategy())

            current_probabilities = distribution_calculator.calculate_approximate_distribution(
                samples_number=self._number_of_samples_for_estimated_distribution_calculation)

            if len(probabilities_list) == 0:
                probabilities_list = [[] for _ in range(len(current_probabilities))]

            for j in range(len(current_probabilities)):
                probabilities_list[j].append(current_probabilities[j])

        random_outcome_index = randint(0, len(current_probabilities))
        self.assertAlmostEqual(number_of_outcomes ** (-1), average(probabilities_list[random_outcome_index]),
                               delta=error_bound)



__author__ = "Tomasz Rybotycki"

"""
Throughout the Boson Sampling Simulation methods there are many different methods of calculating effective scattering
matrix permanent (that is related to samples probabilities). Given that in this project there already are more than 1
of them implemented, we'd like to ensure, that all of them give the same results.

For test purposes I assume that number of observed modes is m = 4 and number of bosons n varies in the tests.
Note that increasing or decreasing m will require adequate updates of input and output states in test cases. 
"""

import unittest
from typing import List

from numpy import allclose, array

from ..theboss.boson_sampling_utilities.permanent_calculators.chin_huh_permanent_calculator import \
    ChinHuhPermanentCalculator
from ..theboss.boson_sampling_utilities.permanent_calculators.classic_permanent_calculator import \
    ClassicPermanentCalculator
from ..theboss.boson_sampling_utilities.permanent_calculators.glynn_gray_permanent_calculator import \
    GlynnGrayPermanentCalculator
from ..theboss.boson_sampling_utilities.permanent_calculators.parallel_chin_huh_permanent_calculator import \
    ParallelChinHuhPermanentCalculator
from ..theboss.boson_sampling_utilities.permanent_calculators.ryser_guan_permanent_calculator import \
    RyserGuanPermanentCalculator
from ..theboss.boson_sampling_utilities.permanent_calculators.ryser_permanent_calculator import \
    RyserPermanentCalculator
from ..theboss.quantum_computations_utilities import generate_haar_random_unitary_matrix


class TestEffectiveScatteringMatrixPermanentsCalculators(unittest.TestCase):

    def setUp(self) -> None:
        self._matrix = generate_haar_random_unitary_matrix(4)

        self._cl_permanent_calculator = ClassicPermanentCalculator(
            matrix=self._matrix, input_state=array([]), output_state=array([])
        )

        self._pch_permanent_calculator = ParallelChinHuhPermanentCalculator(
            matrix=self._matrix, input_state=array([]), output_state=array([])
        )

        self._r_permanent_calculator = RyserPermanentCalculator(
            matrix=self._matrix, input_state=array([]), output_state=array([])
        )

        self._rg_permanent_calculator = RyserGuanPermanentCalculator(
            matrix=self._matrix, input_state=array([]), output_state=array([])
        )

        self._ch_permanent_calculator = ChinHuhPermanentCalculator(
            matrix=self._matrix, input_state=array([]), output_state=array([])
        )

        self._gg_permanent_calculator = GlynnGrayPermanentCalculator(
            matrix=self._matrix, input_state=array([]), output_state=array([])
        )

    def _set_input_and_output_states(self, input_state: List[int],
                                     output_state: List[int]) -> None:
        self._cl_permanent_calculator.input_state = input_state
        self._cl_permanent_calculator.output_state = output_state

        self._pch_permanent_calculator.input_state = input_state
        self._pch_permanent_calculator.output_state = output_state

        self._r_permanent_calculator.input_state = input_state
        self._r_permanent_calculator.output_state = output_state

        self._rg_permanent_calculator.input_state = input_state
        self._rg_permanent_calculator.output_state = output_state

        self._ch_permanent_calculator.input_state = input_state
        self._ch_permanent_calculator.output_state = output_state

        self._r_permanent_calculator.input_state = input_state
        self._r_permanent_calculator.output_state = output_state

        self._gg_permanent_calculator.input_state = input_state
        self._gg_permanent_calculator.output_state = output_state

    def _do_all_assertions(self) -> None:
        cl_permanent = self._cl_permanent_calculator.compute_permanent()

        self.assertTrue(
            allclose(cl_permanent, self._gg_permanent_calculator.compute_permanent()))

        self.assertTrue(
            allclose(cl_permanent, self._r_permanent_calculator.compute_permanent()))

        self.assertTrue(
            allclose(cl_permanent, self._rg_permanent_calculator.compute_permanent()))

        self.assertTrue(
            allclose(cl_permanent, self._ch_permanent_calculator.compute_permanent()))

    def test_full_input_output_case(self) -> None:
        self._set_input_and_output_states([1, 1, 1, 1], [1, 1, 1, 1])
        self._do_all_assertions()

    def test_not_full_input_output_case(self) -> None:
        self._set_input_and_output_states([1, 1, 1, 0], [1, 0, 1, 1])
        self._do_all_assertions()

    def test_binned_input_case(self) -> None:
        self._set_input_and_output_states([2, 1, 0, 0], [0, 1, 1, 1])
        self._do_all_assertions()

    def test_binned_output_case(self) -> None:
        self._set_input_and_output_states([1, 1, 1, 0], [2, 1, 0, 0])
        self._do_all_assertions()

    def test_binned_input_binned_output_case(self) -> None:
        self._set_input_and_output_states([2, 1, 1, 0], [1, 1, 0, 2])
        self._do_all_assertions()

    def test_skipping_input_case(self) -> None:
        self._set_input_and_output_states([0, 0, 4, 0], [1, 1, 0, 2])
        self._do_all_assertions()

    def test_skipping_output_case(self) -> None:
        self._set_input_and_output_states([1, 1, 0, 2], [0, 1, 3, 0])
        self._do_all_assertions()

    def test_skipping_input_skipping_output_case(self) -> None:
        self._set_input_and_output_states([0, 2, 0, 2], [0, 0, 4, 0])
        self._do_all_assertions()

    def test_edge_case_no_particles(self) -> None:
        self._set_input_and_output_states([0, 0, 0, 0], [0, 0, 0, 0])
        self._do_all_assertions()

    def test_parallel_binned_input_binned_output_case(self) -> None:
        """
            TR: Given that the only difference between Chin-Huh and Parallel Chin-Huh is
                the parallelization in the main method, the only thing I believe I need
                to check is if it works. Results are expected to be the same in every
                case.
        """
        self._set_input_and_output_states([2, 1, 1, 0], [1, 1, 0, 2])
        self.assertTrue(allclose(self._cl_permanent_calculator.compute_permanent(),
                                 self._pch_permanent_calculator.compute_permanent()))



__author__ = "Tomasz Rybotycki"

import unittest

from numpy import array, complex128, int64, allclose

from ..theboss.boson_sampling_utilities.permanent_calculators.bs_permanent_calculator_factory import \
    BSPermanentCalculatorFactory, PermanentCalculatorType
from ..theboss.distribution_calculators.bs_exact_distribution_with_uniform_losses import \
    BSDistributionCalculatorWithUniformLosses, BSDistributionCalculatorWithFixedLosses, \
    BosonSamplingExperimentConfiguration


class TestExactLossyBosonSamplingDistributionCalculator(unittest.TestCase):

    def setUp(self) -> None:
        # Define some additional variables for more clear experiment configuration assignment.
        self.number_of_particles_lost = 2

        # Generate permutation matrix and define initial state.
        self.permutation_matrix = array([
            [0, 0, 1, 0, 0],
            [1, 0, 0, 0, 0],
            [0, 0, 0, 1, 0],
            [0, 0, 0, 0, 1],
            [0, 1, 0, 0, 0],
        ], dtype=complex128)

        self.initial_state = array([1, 1, 1, 0, 0], dtype=int64)

        # Create configuration object.
        self.experiment_configuration = BosonSamplingExperimentConfiguration(
            interferometer_matrix=self.permutation_matrix,
            initial_state=self.initial_state,
            number_of_modes=len(self.initial_state),
            initial_number_of_particles=sum(self.initial_state),
            number_of_particles_lost=self.number_of_particles_lost,
            number_of_particles_left=sum(
                self.initial_state) - self.number_of_particles_lost,
            uniform_transmissivity=0.8
        )

        self._calculator_type = PermanentCalculatorType.PARALLEL_CHIN_HUH
        self._permanent_calculator_factory = BSPermanentCalculatorFactory(matrix=None,
                                                                          input_state=None,
                                                                          output_state=None,
                                                                          calculator_type=self._calculator_type)
        self._permanent_calculator = self._permanent_calculator_factory.generate_calculator()

    def test_probabilities_sum_in_fixed_losses_scenario(self) -> None:
        exact_distribution_calculator = \
            BSDistributionCalculatorWithFixedLosses(self.experiment_configuration,
                                                    self._permanent_calculator)
        exact_distribution = exact_distribution_calculator.calculate_distribution()
        self.assertAlmostEqual(sum(exact_distribution), 1.0, delta=1e-4)

    def test_probabilities_sum_in_uniform_losses_scenario(self) -> None:
        exact_distribution_calculator = \
            BSDistributionCalculatorWithUniformLosses(self.experiment_configuration,
                                                      self._permanent_calculator)
        exact_distribution = exact_distribution_calculator.calculate_distribution()
        self.assertAlmostEqual(sum(exact_distribution), 1.0, delta=1e-4)

    def test_probabilities(self) -> None:
        exact_distribution_calculator = \
            BSDistributionCalculatorWithUniformLosses(self.experiment_configuration,
                                                      self._permanent_calculator)
        exact_distribution = exact_distribution_calculator.calculate_distribution()
        self.assertTrue(
            allclose(
                exact_distribution,
                [
                    0.008, 0.032, 0.0, 0.0, 0.032, 0.032, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                    0.128, 0.0, 0.0, 0.0, 0.128, 0.0, 0.0, 0.128, 0.0, 0.0, 0.0, 0.0,
                    0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                    0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.512, 0.0, 0.0,
                    0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                ],
            )
        )
